/* -*-c++-*-
 * Delta3D Open Source Game and Simulation Engine
 * Copyright (C) 2005, BMH Associates, Inc.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Matthew W. Campbell
 * David Guthrie
 */

#ifndef DELTA_NAMED_GROUP_PARAMETER
#define DELTA_NAMED_GROUP_PARAMETER

#include <dtDAL/namedparameter.h>

#include <dtCore/refptr.h>

#include <map>
#include <vector>

namespace dtDAL
{
   class DT_DAL_EXPORT NamedGroupParameter : public NamedParameter
   {
      public:
         typedef std::map<dtUtil::RefString, dtCore::RefPtr<NamedParameter> > ParameterList;

         NamedGroupParameter(const dtUtil::RefString& name);

         virtual void ToDataStream(dtUtil::DataStream& stream) const;

         virtual bool FromDataStream(dtUtil::DataStream& stream);

         virtual const std::string ToString() const;

         virtual bool FromString(const std::string& value);

         virtual void CopyFrom(const NamedParameter& otherParam);

         /**
          * Adds a parameter to the group
          * @param name The name of the parameter to add
          * @param type The type of parameter it is, corresponding with dtDAL::DataType
          * @return A pointer to the parameter
          * @see dtDAL::DataType
          * @throws dtUtil::Exception if the name specified is already used.
          */
         NamedParameter* AddParameter(const dtUtil::RefString& name, dtDAL::DataType& type);

         /**
          * Removes a parameter to the group
          * @param name The name of the parameter to remove
          * @return A ref pointer to the parameter removed or NULL if nothing was removed.  It returns a ref ptr in case
          *         this group is the only object holding on to a reference.
          */
         dtCore::RefPtr<NamedParameter> RemoveParameter(const dtUtil::RefString& name);

         /**
          * Adds a parameter to the group
          * @param param the new parameter to add
          * @see dtDAL::DataType
          * @throws dtUtil::Exception if the name specified is already used.
          */
         void AddParameter(NamedParameter& newParam);

         /**
          * Retrieves the parameter for this group parameter with the given name.
          * @param name The name of the parameter to retrieve
          * @return A pointer to the parameter or NULL if no such parameter exists
          */
         NamedParameter* GetParameter(const dtUtil::RefString& name);

         /**
          * Retrieves const pointer to the parameter for this group parameter with the given name.
          * @param name The name of the parameter to retrieve
          * @return A const pointer to the parameter or NULL if no such parameter exists
          */
         const NamedParameter* GetParameter(const dtUtil::RefString& name) const;

         /**
          * Retrieves all of the parameters in this group.
          * @param toFill The vector to fill with the parameters
          */
         void GetParameters(std::vector<NamedParameter*>& toFill);

         /**
          * Retrieves all of the parameters in this group as const.
          * @param toFill The vector to fill with the parameters
          */
         void GetParameters(std::vector<const NamedParameter*>& toFill) const;

         template <class UnaryFunction>
         void ForEachParameter(UnaryFunction function) const
         {
            /*std::for_each(mParameterList.begin(), mParameterList.end(),
                  std::compose1(std::select2nd<std::pair<std::string, dtCore::RefPtr<NamedParameter> > >(),
                        function));*/
            ParameterList::const_iterator i = mParameterList.begin();
            ParameterList::const_iterator end = mParameterList.end();
            for (; i != end; ++i)
            {
               function(i->second);
            }
         }

         /**
          * Return the amount of iters in the group list
          * @return the size of the internal messageparam map size
          */
         unsigned int GetParameterCount() const;

         /**
          * Sets the message parameter's value from the actor property's value
          */
         virtual void SetFromProperty(const dtDAL::ActorProperty& property);

         /**
          * Sets the actor property's value from the message parameter's value
          */
         virtual void ApplyValueToProperty(dtDAL::ActorProperty& property) const;

         virtual bool operator==(const NamedParameter& toCompare) const;

         /**
          * Creates a message parameter that will hold the given type.
          * @param type the datatype the parameter should hold.
          * @param name the name of the parameter to create.
          * @throw dtUtil::Exception with dtGame::ExceptionEnum::INVALID_PARAMETER if the type is unsupported.
          */
         static dtCore::RefPtr<NamedParameter> CreateFromType(
                  dtDAL::DataType& type,
                  const dtUtil::RefString& name, bool isList = false);

      private:
         ParameterList mParameterList;
   };
}

#endif //DELTA_NAMED_GROUP_PARAMETER
