/* -*-c++-*-
 * Delta3D Open Source Game and Simulation Engine
 * Copyright (C) 2005, BMH Associates, Inc.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Eric Heine
 */
#include <prefix/dtdalprefix.h>
#include <dtDAL/actoractorproperty.h>
#include <dtDAL/map.h>
#include <dtDAL/project.h>

namespace dtDAL
{
   ////////////////////////////////////////////////////////////////////////////
   ActorActorProperty::ActorActorProperty(ActorProxy& actorProxy,
                     const dtUtil::RefString& name,
                     const dtUtil::RefString& label,
                     SetFuncType Set,
                     GetFuncType Get,
                     const dtUtil::RefString& desiredActorClass,
                     const dtUtil::RefString& desc,
                     const dtUtil::RefString& groupName)
      : ActorProperty(DataType::ACTOR, name, label, desc, groupName)
      , mProxy(&actorProxy)
      , SetPropFunctor(Set)
      , GetActorFunctor(Get)
      , mDesiredActorClass(desiredActorClass)
   {

   }

   ////////////////////////////////////////////////////////////////////////////
   void ActorActorProperty::CopyFrom(const ActorProperty& otherProp)
   {
      if (GetDataType() != otherProp.GetDataType())
      {
         LOG_ERROR("Property types are incompatible. Cannot make copy.");
         return;
      }

      const ActorActorProperty& prop =
         static_cast<const ActorActorProperty& >(otherProp);

      SetValue(prop.GetValue());
   }

   ////////////////////////////////////////////////////////////////////////////
   void ActorActorProperty::SetValue(ActorProxy* value)
   {
      if (IsReadOnly())
      {
         LOG_WARNING("SetValue has been called on a property that is read only.");
         return;
      }

      SetPropFunctor(value);
   }

   ////////////////////////////////////////////////////////////////////////////
   ActorProxy* ActorActorProperty::GetValue() const
   {
      if (!mProxy) return NULL;

      return mProxy->GetLinkedActor(GetName());
   }

   ////////////////////////////////////////////////////////////////////////////
   dtCore::DeltaDrawable* ActorActorProperty::GetRealActor()
   {
      return GetActorFunctor();
   }

   ////////////////////////////////////////////////////////////////////////////
   const dtCore::DeltaDrawable* ActorActorProperty::GetRealActor() const
   {
      return GetActorFunctor();
   }

   ////////////////////////////////////////////////////////////////////////////
   bool ActorActorProperty::FromString(const std::string& value)
   {
      if (IsReadOnly())
         return false;

      if (value.empty() || value == "NULL")
      {
         SetValue(NULL);
         return true;
      }

      dtCore::UniqueId newIdValue(value);
      try
      {
         Map* map = Project::GetInstance().GetMapForActorProxy(*mProxy);

         if (map == NULL)
         {
            dtUtil::Log::GetInstance("enginepropertytypes.cpp").LogMessage(dtUtil::Log::LOG_INFO,
                                                                           __FUNCTION__, __LINE__, "Actor does not exist in a map.  Setting property %s with string value failed.",
                                                                           GetName().c_str());
            return false;
         }

         ActorProxy* newProxyValue = map->GetProxyById(newIdValue);
         if (newProxyValue == NULL)
         {
            dtUtil::Log::GetInstance("enginepropertytypes.cpp").LogMessage(dtUtil::Log::LOG_INFO,
                                                                           __FUNCTION__, __LINE__, "Actor with ID %s not found.  Setting property %s with string value failed.",
                                                                           value.c_str(), GetName().c_str());
            return false;
         }

         SetValue(newProxyValue);
         return true;
      }
      catch (const dtDAL::ProjectInvalidContextException& ex)
      {
         dtUtil::Log::GetInstance("enginepropertytypes.cpp").LogMessage(dtUtil::Log::LOG_WARNING,
            __FUNCTION__, __LINE__, "Project context is not set, unable to lookup actors.  Setting property %s with string value failed. Error Message %s.",
            GetName().c_str(), ex.What().c_str());

      }
      catch (const dtUtil::Exception& ex)
      {
         dtUtil::Log::GetInstance("enginepropertytypes.cpp").LogMessage(dtUtil::Log::LOG_WARNING,
            __FUNCTION__, __LINE__, "Error setting ActorActorProperty.  Setting property %s with string value failed. Error Message %s.",
            GetName().c_str(), ex.What().c_str());
      }

      return false;
   }

   ////////////////////////////////////////////////////////////////////////////
   const std::string ActorActorProperty::ToString() const
   {
      return GetValue() == NULL ? "" : GetValue()->GetId().ToString();
   }

   ////////////////////////////////////////////////////////////////////////////////
   const std::string& ActorActorProperty::GetDesiredActorClass() const
   {
      return mDesiredActorClass;
   }

   ////////////////////////////////////////////////////////////////////////////

}
