/*
 * Delta3D Open Source Game and Simulation Engine
 * Copyright (C) 2005, BMH Associates, Inc.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * David Guthrie
 */

#include <prefix/dtutilprefix.h>
#include <dtUtil/mswinmacros.h>
#include <dtUtil/macros.h> //for DTUNREFERENCED_PARAMETER
#include <osg/Version>
#include <algorithm>

#ifdef DELTA_WIN32
#   include <dtUtil/mswin.h>
#   include <direct.h>
#   include <io.h>
#   include <errno.h>
#   define stat64 _stati64
#   define mkdir(x,y) _mkdir((x))

#   if defined(_MT) || defined(_DLL)
_CRTIMP extern int * __cdecl _errno(void);
#      define errno   (*_errno())
#   else   // ndef _MT && ndef _DLL
_CRTIMP extern int errno;
#   endif  // _MT || _DLL

//Linux and Mac OS X
#else
#   include <sys/param.h>
#   include <errno.h>
#endif


#include <osg/Notify>
#include <stack>

#include <dtUtil/fileutils.h>
#include <dtUtil/exception.h>
#include <dtUtil/log.h>

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <iostream>

#include <osgDB/FileUtils>
#include <osgDB/FileNameUtils>

//we only want to NOT use stat64, if it's not defined.
#ifndef stat64
#define stat64 stat
#endif


#ifndef S_ISREG
#define S_ISREG(x) (((x) & S_IFMT) == S_IFREG)
#endif

#ifndef S_ISDIR
#define S_ISDIR(x) (((x) & S_IFMT) == S_IFDIR)
#endif

#ifdef MAX_PATH
   #undef MAX_PATH
#endif

#define MAX_PATH 1024

#include <stack>
#include <queue>


namespace dtUtil
{

   IMPLEMENT_ENUM(FileExceptionEnum)

   FileExceptionEnum FileExceptionEnum::IOException("File IO Exception");
   FileExceptionEnum FileExceptionEnum::FileNotFound("File Not Found");

   dtCore::RefPtr<FileUtils> FileUtils::mInstance;

#ifdef DELTA_WIN32
   const char FileUtils::PATH_SEPARATOR = '\\';
#else
   const char FileUtils::PATH_SEPARATOR = '/';
#endif

   // temporary copy of osgDB::makeDirectory because of some bugs in it.
   bool iMakeDirectory(const std::string& path)
   {
      if (path.empty())
      {
         osg::notify(osg::DEBUG_INFO) << "osgDB::makeDirectory():  cannot create an empty directory" << std::endl;
         return false;
      }

      struct stat64 stbuf;
      if (stat64(path.c_str(), &stbuf) == 0)
      {
         if (S_ISDIR(stbuf.st_mode))
         {
            return true;
         }
         else
         {
            osg::notify(osg::DEBUG_INFO) << "osgDB::makeDirectory():  "  <<
               path << " already exists and is not a  directory!" << std::endl;
            return false;
         }
      }

      std::string dir = path;
      std::stack<std::string> paths;
      while (true)
      {
         if (dir.empty())
         {
            break;
         }

         if (stat64(dir.c_str(), &stbuf) < 0)
         {
            switch (errno)
            {
            case ENOENT:
            case ENOTDIR:
               paths.push(dir);
               break;

            default:
               osg::notify(osg::DEBUG_INFO) <<  "osgDB::makeDirectory(): "  << strerror(errno) << std::endl;
               return false;
            }
         }
         else
         {
            break;
         }
         dir = osgDB::getFilePath(std::string(dir));
      }

      while (!paths.empty())
      {
         std::string dir = paths.top();

         if (mkdir(dir.c_str(), 0755) < 0)
         {
            osg::notify(osg::DEBUG_INFO) << "osgDB::makeDirectory():  "  << strerror(errno) << std::endl;
            return false;
         }
         paths.pop();
      }
      return true;
   }

   //-----------------------------------------------------------------------
   bool FileUtils::FileExists(const std::string& strFile) const
   {
      return GetFileInfo(strFile).fileType != FILE_NOT_FOUND;
   }

   //-----------------------------------------------------------------------
   void FileUtils::FileCopy(const std::string& strSrc, const std::string& strDest, bool bOverwrite) const
   {
      FILE* pSrcFile;
      FILE* pDestFile;

      struct stat tagStat;

      // Make absolutely certain these two strings don't point to the same file.
      if (!IsSameFile(strSrc, strDest))
      {
         if (!FileExists(strSrc))
         {
            throw dtUtil::FileNotFoundException(
                   std::string("Source file does not exist: \"") + strSrc + "\"", __FILE__, __LINE__);
         }

         // Open the source file for reading.
         pSrcFile = fopen(strSrc.c_str(), "rb");
         if (pSrcFile == NULL)
         {
            throw dtUtil::FileUtilIOException(
                   std::string("Unable to open source file for reading: \"") + strSrc + "\"", __FILE__, __LINE__);
         }

         try
         {
            if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
            {
               mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Source file exists.");
            }

            std::string destFile = strDest;

            FileType ft = GetFileInfo(strDest).fileType;

            // Check to see if the destination is a file or directory.
            if (ft == DIRECTORY)
            {
               if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
               {
                  mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Destination is a directory.");
               }

               // If the file is a directory, append the base name of the source file to the destination
               // to make the new file name.
               if (strDest[strDest.size()-1] != FileUtils::PATH_SEPARATOR)
               {
                  destFile = strDest + FileUtils::PATH_SEPARATOR + osgDB::getSimpleFileName(strSrc);
               }
               else
               {
                  destFile = strDest + osgDB::getSimpleFileName(strSrc);
               }
            }
            else
            {
               if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
               {
                  mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Destination is a file.");
               }
            }

            if (FileExists(destFile) && !bOverwrite)
            {
               throw dtUtil::FileUtilIOException(
                      std::string("Destination file exists, but overwriting is turned off: \"") + destFile + "\"", __FILE__, __LINE__);
            }

            pDestFile = fopen(destFile.c_str(), "wb");

            if (pDestFile == NULL)
            {
               // make sure to close the source file.
               throw dtUtil::FileUtilIOException(
                      std::string("Unable to open destination for writing: \"") + destFile + "\"", __FILE__, __LINE__);
            }

            try
            {
               if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
               {
                  mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, "Destination opened for reading.");
               }

               stat(strSrc.c_str(), &tagStat);
               long i = 0;
               char buffer[4096];
               while (i < tagStat.st_size)
               {
                  size_t readCount = fread(buffer, 1, 4096, pSrcFile);
                  if (readCount > 0)
                  {
                     size_t numWritten = fwrite(buffer, 1, readCount, pDestFile);
                     if (numWritten < readCount)
                     {
                        throw dtUtil::FileUtilIOException(
                               std::string("Unable to write to destinate file: \"") + destFile + "\"", __FILE__, __LINE__);
                     }
                     i += readCount;
                  }
               }
               fclose(pDestFile);
               fclose(pSrcFile);
            }
            catch (dtUtil::Exception& ex1)
            {
               fclose(pDestFile);
               throw ex1;
            }
         }
         catch (dtUtil::Exception& ex)
         {
            fclose(pSrcFile);
            throw ex;
         }

      }
      // if the source equals the destination, this method is really a noop.
      // (Not to mention the fact that if you attempt to copy a file onto itself
      // in this manner then you will end up blowing it away).

   }

   //-----------------------------------------------------------------------
   void FileUtils::FileMove(const std::string& strSrc, const std::string& strDest, bool bOverwrite) const
   {
      if (GetFileInfo(strSrc).fileType != REGULAR_FILE)
      {
         throw dtUtil::FileNotFoundException(
                std::string("Source file was not found or is a Directory: \"") + strSrc + "\"", __FILE__, __LINE__);
      }

      FileType ft = GetFileInfo(strDest).fileType;

      std::string destFile = strDest;

      // Check to see if the destination is a directory.
      if (ft == DIRECTORY)
      {
         if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
         {
            mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Destination is a directory.");
         }

         // Check to see if the destination is a file or directory.
         if (strDest[strDest.size()-1] != FileUtils::PATH_SEPARATOR)
         {
            destFile = strDest + FileUtils::PATH_SEPARATOR + osgDB::getSimpleFileName(strSrc);
         }
         else
         {
            destFile = strDest + osgDB::getSimpleFileName(strSrc);
         }

         ft = GetFileInfo(destFile).fileType;
      }
      else
      {
         if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
         {
            mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Destination is a file.");
         }
      }

      if (ft != FILE_NOT_FOUND && !bOverwrite)
      {
         throw dtUtil::FileUtilIOException(
                std::string("Destination file exists and the call was not set to overwrite: \"") + strDest + "\"", __FILE__, __LINE__);
      }


      // first check to see if the file can be moved without copying it.
      if (rename(strSrc.c_str(), destFile.c_str()) == 0)
      {
         return;
      }

      if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
      {
         mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, "Rename failed, attempting to copy file and delete the source");
      }

      // copy the file
      FileCopy(strSrc, strDest, bOverwrite);

      // attempt to delete the original file.
      if (unlink(strSrc.c_str()) != 0)
      {
         throw dtUtil::FileUtilIOException(
                std::string("Unable to delete \"") + strSrc + "\" but file copied to new location.", __FILE__, __LINE__);
      }
   }

   //-----------------------------------------------------------------------
   void FileUtils::FileDelete(const std::string& strFile) const
   {
      FileType ft = GetFileInfo(strFile).fileType;

      // If the file does not exist, then ignore.
      if (ft == FILE_NOT_FOUND)
      {
         return;
      }

      if (ft != REGULAR_FILE)
      {
         throw dtUtil::FileUtilIOException(
                std::string("File \"") + strFile + "\" is a directory.", __FILE__, __LINE__);
      }

      if (unlink(strFile.c_str()) != 0)
      {
         throw dtUtil::FileUtilIOException(
                std::string("Unable to delete \"") + strFile + "\".", __FILE__, __LINE__);
      }
   }

   //-----------------------------------------------------------------------
   const struct FileInfo FileUtils::GetFileInfo(const std::string& strFile) const
   {
      struct FileInfo info;
      struct stat tagStat;

      // chop trailing slashes off
      std::string choppedStr = strFile;
      if (strFile.size() > 0 && (strFile[strFile.size() - 1] == '\\' ||
                                 strFile[strFile.size() - 1] == '/'))
      {
         choppedStr = strFile.substr(0, strFile.length() - 1);
      }

      if (stat(choppedStr.c_str(), &tagStat) != 0)
      {
         // throw dtUtil::FileNotFoundException( std::string("Cannot open file ") + choppedStr);
         info.fileType = FILE_NOT_FOUND;
         return info;
      }

      info.extensionlessFileName = osgDB::getStrippedName(choppedStr);
      info.baseName = osgDB::getSimpleFileName(choppedStr);
      info.fileName = choppedStr;
      info.path = osgDB::getFilePath(choppedStr);
      info.extension = osgDB::getFileExtension(choppedStr);

      info.size = tagStat.st_size;
      info.lastModified = tagStat.st_mtime;

      if (S_ISDIR(tagStat.st_mode))
      {
         info.fileType = DIRECTORY;
      }
      else
      {
         // Anything else is a regular file, including special files
         // this is incomplete, but not a case that we deemed necessary to handle.
         // Symbolic links should NOT show up because stat was called, not lstat.
         info.fileType = REGULAR_FILE;
      }

      return info;
   }

   //-----------------------------------------------------------------------
   void FileUtils::CleanupFileString(std::string& strFileOrDir)
   {
      if (strFileOrDir.empty())
      {
         return;
      }

      // convert all separators to unix-style for conformity
      for (unsigned int i = 0; i < strFileOrDir.length(); ++i)
      {
         if (strFileOrDir[i] == '\\')
         {
            strFileOrDir[i] = '/';
         }
      }

      // get rid of trailing separators
      if (strFileOrDir[strFileOrDir.length()-1] == '/')
      {
         strFileOrDir = strFileOrDir.substr(0, strFileOrDir.length()-1);
      }
   }

   //-----------------------------------------------------------------------
   bool FileUtils::IsAbsolutePath(std::string strFileOrDir)
   {
      // just in case, make sure we are using a valid path
      CleanupFileString(strFileOrDir);

      dtUtil::FileInfo info = GetFileInfo(strFileOrDir);
      if (info.fileType == dtUtil::FILE_NOT_FOUND)
      {
         return false;
      }

      // handle unix-style paths
      if (strFileOrDir.length() > 0 && strFileOrDir[0] == '/')
      {
         return true;
      }

      // handle windows-style paths
      else if (strFileOrDir.length() > 1 && strFileOrDir[1] == ':')
      {
         return true;
      }

      return false;
   }

   //-----------------------------------------------------------------------
   void FileUtils::MakeDirectoryEX(std::string strDir)
   {
      if (strDir.empty())
      {
         return;
      }

      // first make sure the directory string is clean
      CleanupFileString(strDir);

      // this generated string is used to check each subdirectory in the path
      std::string dir;

      // prepare tokenizer
      char* buffer = new char[strDir.length()+1];
      strcpy(buffer, strDir.c_str());
      char* tok = strtok(buffer, "/");

      // setup root directory
      if (strDir[0] == '/')
      {
         dir = "/";
         dir += tok;
      }
      else
      {
         dir = tok;
         dir += "/";
      }

      // traverse each subdirectory in path and create directories as needed
      while (tok)
      {
         // debug
         //std::cout << tok << std::endl;

         // create directory if necessary
         if (!DirExists(dir))
         {
            MakeDirectory(dir);
         }

         // get next subdirectory
         tok = strtok(NULL, "/");
         if (tok)
         {
            if (dir[dir.length()-1] != '/') { dir += '/'; }
            dir += tok;
         }
      }

      // cleanup
      delete [] buffer;
   }

   //-----------------------------------------------------------------------
   void FileUtils::ChangeDirectory(const std::string& path)
   {
      ChangeDirectoryInternal(path);
      mStackOfDirectories.clear();
   }

   //-----------------------------------------------------------------------
   const std::string& FileUtils::CurrentDirectory() const
   {
      return mCurrentDirectory;
   }

   //-----------------------------------------------------------------------
   void FileUtils::ChangeDirectoryInternal(const std::string& path)
   {
      if (chdir(path.c_str()) == -1)
      {
         throw dtUtil::FileNotFoundException( std::string("Cannot open directory ") + path, __FILE__, __LINE__);
      }
      char buf[512];
      char* bufAddress = getcwd(buf, 512);
      if (buf != bufAddress)
      {
         throw dtUtil::FileUtilIOException( std::string("Cannot get current working directory"), __FILE__, __LINE__);
      }

      mCurrentDirectory = buf;

      if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
      {
         std::string message("Changed directory to \"");
         message += mCurrentDirectory;
         message += "\".";
         mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, message.c_str());
      }
   }

   //-----------------------------------------------------------------------
   void FileUtils::PushDirectory(const std::string& path)
   {
      if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
      {
         std::string message("Pushing Directory \"");
         message += path;
         message += "\".";
         mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, message.c_str());
      }
      std::string old = mCurrentDirectory;
      ChangeDirectoryInternal(path);
      mStackOfDirectories.push_back(old);
   }

   //-----------------------------------------------------------------------
   void FileUtils::PopDirectory()
   {
      if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
      {
         mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Popping Directory.");
      }

      if (mStackOfDirectories.empty())
      {
         return;
      }

      ChangeDirectoryInternal(mStackOfDirectories.back());
      mStackOfDirectories.pop_back();
   }

   //-----------------------------------------------------------------------
   const std::string FileUtils::GetAbsolutePath(const std::string& relativePath) const
   {
      std::string result;
      std::string old = mCurrentDirectory;
      try
      {
         if (chdir(relativePath.c_str()) == -1)
         {
            throw dtUtil::FileNotFoundException( std::string("Cannot open directory ") + relativePath, __FILE__, __LINE__);
         }
         char buf[512];
         char* bufAddress = getcwd(buf, 512);
         if (bufAddress != buf)
         {
            throw dtUtil::FileUtilIOException( std::string("Cannot get current working directory"), __FILE__, __LINE__);
         }
         result = buf;
      }
      catch (const dtUtil::Exception& ex)
      {
         if (chdir(old.c_str()) == -1)
         {
            mLogger->LogMessage(dtUtil::Log::LOG_ERROR, __FUNCTION__, __LINE__,
                                "Attempting to reset current directory to \"%s\", but an error occured doing so.", old.c_str());
         }
         throw ex;
      }
      if (chdir(old.c_str()) == -1)
      {
         mLogger->LogMessage(dtUtil::Log::LOG_ERROR, __FUNCTION__, __LINE__,
                             "Attempting to reset current directory to \"%s\", but an error occured doing so.", old.c_str());
      }
      return result;
   }

   //////////////////////////////////////////////////////////////////////////
   std::string getFileExtensionIncludingDot(const std::string& fileName)
   {
#if defined(OSG_VERSION_MAJOR) && defined(OSG_VERSION_MINOR) && OSG_VERSION_MAJOR <= 2  && OSG_VERSION_MINOR <= 6
      //copied from osgDB/FileNameUtils.cpp (v2.8.0)
      std::string::size_type dot = fileName.find_last_of('.');
      if (dot==std::string::npos) return std::string("");
      return std::string(fileName.begin()+dot,fileName.end());
#else
      return osgDB::getFileExtensionIncludingDot(fileName);
#endif
   }

   ////////////////////////////////////////////////////////////////////////////////
   class ToLowerClass
   {
   public:
      ToLowerClass(){}

      char operator() (char& elem) const
      {
         return tolower(elem);
      }
   };

   //////////////////////////////////////////////////////////////////////////
   DirectoryContents FileUtils::DirGetFiles(const std::string& path,
                                            const FileExtensionList& extensions) const
   {
      FileInfo ft = GetFileInfo(path);
      if (ft.fileType == FILE_NOT_FOUND)
      {
         throw dtUtil::FileNotFoundException(
                std::string("Path not Found: \"") + path + "\".", __FILE__, __LINE__);
      }
      else if (ft.fileType == REGULAR_FILE)
      {
         throw dtUtil::FileUtilIOException(
                std::string("Path does not specify a directory: \"") + path + "\".", __FILE__, __LINE__);
      }

      DirectoryContents dirContents = osgDB::getDirectoryContents(path);
      if (extensions.empty())
      {
         return dirContents;
      }

      DirectoryContents filteredContents;
      // iterate over contents, looking for files that have the requested extensions
      DirectoryContents::iterator dirItr = dirContents.begin();
      while (dirItr != dirContents.end())
      {
         FileExtensionList::const_iterator extItr = extensions.begin();
         while (extItr != extensions.end())
         {
            std::string testExt = getFileExtensionIncludingDot((*dirItr));
            std::transform(testExt.begin(), testExt.end(), testExt.begin(), ToLowerClass());
            std::string validExt = (*extItr);
            std::transform(validExt.begin(), validExt.end(), validExt.begin(), ToLowerClass());
            if (testExt == validExt)
            {
               filteredContents.push_back((*dirItr));
               // stop when we find at least one match, to avoid duplicate file entries, in
               // case extensions contains duplicate entries.
               break;
            }
            ++extItr;
         }
         ++dirItr;
      }

      return filteredContents;
   }

   //-----------------------------------------------------------------------
   DirectoryContents FileUtils::DirGetSubs(const std::string& path) const
   {
      DirectoryContents vec;

      DirectoryContents dirCont = DirGetFiles(path);

      for (DirectoryContents::const_iterator i = dirCont.begin(); i != dirCont.end(); ++i)
      {
         if (GetFileInfo(path + PATH_SEPARATOR + *i).fileType == DIRECTORY && (*i != ".") && (*i != ".."))
         {
            vec.push_back(*i);
         }
      }

      return vec;
   }


   void FileUtils::InternalDirCopy(const std::string& srcPath,
                                   const std::string& destPath, bool bOverwrite) const
   {
      FileType destFileType = GetFileInfo(destPath).fileType;
      //std::cout << "Copying " << srcPath << " to " << destPath << std::endl;

      if (destFileType == REGULAR_FILE)
      {
         throw dtUtil::FileNotFoundException(
                std::string("The destination path must be a directory: \"") + destPath + "\"", __FILE__, __LINE__);
      }

      if (destFileType == FILE_NOT_FOUND)
      {
         MakeDirectory(destPath);
      }

      DirectoryContents contents = DirGetFiles(srcPath);
      for (DirectoryContents::iterator i = contents.begin(); i != contents.end(); ++i)
      {
         if (*i == "." || *i == "..")
         {
            continue;
         }
         const std::string newSrcPath = srcPath + PATH_SEPARATOR + *i;
         const std::string newDestPath = destPath + PATH_SEPARATOR + *i;
         FileInfo fi = GetFileInfo(newSrcPath);
         if (fi.fileType == DIRECTORY)
         {
            if (destPath.size() >= newSrcPath.size() &&
                destPath.substr(0, newSrcPath.size()) == newSrcPath &&
                (destPath.size() == newSrcPath.size() || destPath[newSrcPath.size()] == PATH_SEPARATOR))
            {
               if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_WARNING))
               {
                  mLogger->LogMessage(dtUtil::Log::LOG_WARNING, __FUNCTION__, __LINE__, "Can't copy %s into itself.", newSrcPath.c_str());
               }
            }
            else
            {
               InternalDirCopy(newSrcPath, newDestPath, bOverwrite);
            }
         }
         else
         {
            FileCopy(newSrcPath, newDestPath, bOverwrite);
         }
      }
   }

   //-----------------------------------------------------------------------
   void FileUtils::DirCopy(const std::string& srcPath,
                           const std::string& destPath, bool bOverwrite, bool copyContentsOnly) const
   {
      if (!DirExists(srcPath))
      {
         throw dtUtil::FileNotFoundException(
                std::string("Source directory does not exist: \"") + srcPath + "\"", __FILE__, __LINE__);
      }

      FileType destFileType = GetFileInfo(destPath).fileType;

      if (destFileType == REGULAR_FILE)
      {
         throw dtUtil::FileNotFoundException(
                std::string("The destination path must be a directory: \"") + destPath + "\"", __FILE__, __LINE__);
      }

      bool createDest = destFileType == FILE_NOT_FOUND;

      std::string fullSrcPath = GetAbsolutePath(srcPath);
      // from here, the code can assume srcPath exists.
      if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
      {
         mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Source directory \"%s\" exists.", fullSrcPath.c_str());
      }

      std::string fullDestPath;

      if (createDest)
      {
         MakeDirectory(destPath);
         fullDestPath = GetAbsolutePath(destPath);
         if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
         {
            mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Destination directory \"%s\" has been created.",
                                fullDestPath.c_str());
         }
      }
      else
      {
         fullDestPath = GetAbsolutePath(destPath);
         if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
         {
            mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__, "Destination directory \"%s\" exists.",
                                fullDestPath.c_str());
         }

         if ((copyContentsOnly && fullSrcPath == fullDestPath)
              || (!copyContentsOnly && osgDB::getFilePath(fullSrcPath) == fullDestPath))
         {
            throw dtUtil::FileUtilIOException(
                   std::string("The source equals the destination: \"") + srcPath + "\"", __FILE__, __LINE__);
         }

         if (!copyContentsOnly)
         {
            const std::string& srcName = osgDB::getSimpleFileName(fullSrcPath);
            fullDestPath += PATH_SEPARATOR + srcName;

            if (DirExists(fullDestPath) && !bOverwrite)
            {
               throw dtUtil::FileUtilIOException(
                           std::string("Cannot overwrite directory (overwrite flag is false): \"") + srcPath + "\"",
                           __FILE__, __LINE__);
            }

            if (!DirExists(fullDestPath))
            {
               MakeDirectory(fullDestPath);
            }

            if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
            {
               mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__,
                                   "Destination directory \"%s\" created - copyContentsOnly is false.",
                                   fullDestPath.c_str());
            }
         }
         else
         {
            if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
            {
               mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__,
                                   "Destination directory \"%s\" exists - copyContentsOnly is true.",
                                   fullDestPath.c_str());
            }
         }
      }

      InternalDirCopy(fullSrcPath, fullDestPath, bOverwrite);
   }

   //-----------------------------------------------------------------------
   bool FileUtils::DirDelete(const std::string& strDir, bool bRecursive)
   {
      if (bRecursive)
      {
         if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
         {
            mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__,
                                "Attempting to recursively delete %s.", strDir.c_str());
         }
         try
         {
            PushDirectory(strDir);
            try
            {
               RecursDeleteDir(true);
            }
            catch(const dtUtil::Exception& ex)
            {
               PopDirectory();
               throw ex;
            }

            PopDirectory();
         }
         catch (const dtUtil::FileNotFoundException& ex)
         {
            DTUNREFERENCED_PARAMETER(ex); //squelch "unreferenced" warning

            // if we get a file not found trying to recurse into the top directory
            // then the directory does not exist, so there is no need to throw an exception.
            if (!DirExists(strDir))
            {
               if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
               {
                  mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__,
                     "Directory %s doesn't exist to delete. Ignoring.", strDir.c_str());
               }
               return true;
           }
         }
         catch (const dtUtil::Exception& ex)
         {
            throw ex;
         }
      }
      else
      {
         if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
         {
            mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__,
                                "Attempting to delete %s, but not recursively.", strDir.c_str());
         }
      }

      errno = 0;
      if (rmdir(strDir.c_str()) != 0)
      {
         if (!bRecursive && errno == ENOTEMPTY)
         {
            return false;
         }
         else if (errno == ENOENT)
         {
            if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
            {
               mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__,
                                   "Directory %s doesn't exist to delete. Ignoring.", strDir.c_str());
            }
            return true;
         }
         else
         {
            throw dtUtil::FileUtilIOException(
                   std::string("Unable to delete directory \"") + strDir + "\":" + strerror(errno), __FILE__, __LINE__);
         }
      }

      return true;
   }

   void FileUtils::MakeDirectory(const std::string& strDir) const
   {
      if (!iMakeDirectory(strDir))
      {
         FileType ft = GetFileInfo(strDir).fileType;
         if (ft == REGULAR_FILE)
         {
            throw dtUtil::FileUtilIOException( std::string("Cannot create directory. ")
                   + strDir + " is an existing non-directory file.", __FILE__, __LINE__);
         }
         else if (ft == DIRECTORY)
         {
            return;
         }

         if (!DirExists(osgDB::getFilePath(strDir)))
         {
            throw dtUtil::FileNotFoundException( std::string("Cannot create directory ")
                   + strDir + ". Parent directory doesn't exist.", __FILE__, __LINE__);
         }
         else
         {
            throw dtUtil::FileUtilIOException( std::string("Cannot create directory ") + strDir + ".", __FILE__, __LINE__);
         }
      }
   }


   //-----------------------------------------------------------------------
   bool FileUtils::DirExists(const std::string& strDir) const
   {
      return GetFileInfo(strDir).fileType == DIRECTORY;
   }

   //-----------------------------------------------------------------------
   std::string FileUtils::RelativePath(const std::string& absolutePath, const std::string& file) const
   {
      std::string relativePath;

      for (size_t i = 0; i < file.size(); i++)
      {
         if (file[i] != absolutePath[i])
         {
            if (file[i] == '/')
            {
               relativePath = file.substr(i + 1);
            }
            else
            {
               relativePath = file.substr(i);
            }

            break;
         }
      }

      return relativePath;
   }

   //-----------------------------------------------------------------------
   bool FileUtils::IsSameFile(const std::string& file1, const std::string& file2) const
   {
      // If path names are different, we still could be pointing at the same file
      // on disk.
      struct stat stat1 = {0};
      struct stat stat2 = {0};

      if (stat(file1.c_str(), &stat1) != 0)
      {
         return false;
      }

      if (stat(file2.c_str(), &stat2) != 0)
      {
         return false;
      }

// only Unix variants support inodes
#ifndef WIN32
      if (stat1.st_ino == stat2.st_ino)
      {
         return true;
      }
#else // WIN32 -- No inodes in Windows -- we'll have to imitate an inode's functionality.
      if (stat1.st_atime == stat2.st_atime &&
          stat1.st_ctime == stat2.st_ctime &&
          stat1.st_mtime == stat2.st_mtime &&
          stat1.st_gid   == stat2.st_gid   &&
          stat1.st_uid   == stat2.st_uid   &&
          stat1.st_mode  == stat2.st_mode  &&
          stat1.st_size  == stat2.st_size)
      {
         // also make sure the file names (NOT paths) are the same
         std::string file1Name = file1.substr(file1.find_last_of("\\/"));
         // Windows is case insensitive
         std::transform(file1Name.begin(), file1Name.end(), file1Name.begin(), tolower);
         std::string file2Name = file2.substr(file2.find_last_of("\\/"));
         std::transform(file2Name.begin(), file2Name.end(), file2Name.begin(), tolower);

         if (file1Name == file2Name)
         {
            return true;
         }
      }
#endif

      return false;
   }

   //-----------------------------------------------------------------------
   void FileUtils::RecursDeleteDir(bool bRecursive)
   {
      // this method assumes one is IN the directory that you want to delete.
      DirectoryContents dirCont = DirGetFiles(mCurrentDirectory);

      // iterate over all of the directory contents.
      for (DirectoryContents::const_iterator i = dirCont.begin(); i != dirCont.end(); ++i)
      {
         FileType ft = GetFileInfo(*i).fileType;
         if (ft == REGULAR_FILE)
         {
            // Delete regular files.
            errno = 0;
            if (unlink(i->c_str()) < 0)
            {
               throw dtUtil::FileUtilIOException(
                      std::string("Unable to delete directory \"") + *i + "\":" + strerror(errno), __FILE__, __LINE__);
            }
         }
         else if ((*i != ".") && (*i != "..") && ft == DIRECTORY && bRecursive)
         {
            // if it's a directory and it's not the "." or ".." special directories,
            // change into that directory and recurse.
            ChangeDirectoryInternal(*i);
            RecursDeleteDir(true);
            // now that the directory is empty, remove it.
            errno = 0;
            if (rmdir(i->c_str()) != 0)
            {
               if (errno == ENOENT)
               {
                  if (mLogger->IsLevelEnabled(dtUtil::Log::LOG_DEBUG))
                  {
                     mLogger->LogMessage(dtUtil::Log::LOG_DEBUG, __FUNCTION__, __LINE__,
                                         "Directory %s doesn't exist to delete. Ignoring.", (mCurrentDirectory + PATH_SEPARATOR + *i).c_str());
                  }
                  return;
               }
               else
               {
                  throw dtUtil::FileUtilIOException(
                         std::string("Unable to delete directory \"") + mCurrentDirectory + PATH_SEPARATOR + *i + "\":" + strerror(errno), __FILE__, __LINE__);

               }
            }
         }
      }

      // change up so that when the method ends, the directory is ready to be removed.
      ChangeDirectoryInternal(std::string(".."));
   }

   ////////////////////////////////////////////////////////////////////////////////
   std::string FileUtils::ConcatPaths(const std::string& left, const std::string& right)
   {
#if defined(OPENSCENEGRAPH_MAJOR_VERSION) && OPENSCENEGRAPH_MAJOR_VERSION <= 2 && defined(OPENSCENEGRAPH_MINOR_VERSION) && OPENSCENEGRAPH_MINOR_VERSION <= 8 && defined(OPENSCENEGRAPH_PATCH_VERSION) && OPENSCENEGRAPH_PATCH_VERSION < 3
      if (left.empty())
      {
         return right;
      }
#endif

      return osgDB::concatPaths(left, right);
   }

   //-----------------------------------------------------------------------
   FileUtils::FileUtils()
   {
      mLogger = &dtUtil::Log::GetInstance(std::string("fileutils.cpp"));
      // assign the current directory
      ChangeDirectory(".");
   }

   //-----------------------------------------------------------------------
   FileUtils::~FileUtils() {}

   /*void FileUtils::AbsoluteToRelative(const std::string& pcAbsPath, std::string& relPath)
   {
      char pcRelPath[MAX_PATH];
      char acTmpCurrDir[MAX_PATH];
      char acTmpAbsPath[MAX_PATH];
      int count = 0;
      std::string curDir = CurrentDirectory();

      for (size_t i = 0; i < curDir.size(); ++i)
      {
         if (curDir[i] == '\\')
         {
            curDir[i] = '/';
         }
      }

      strcpy(acTmpCurrDir, curDir.c_str());
      strcpy(acTmpAbsPath, pcAbsPath.c_str());

      std::stack<char*> tmpStackAbsPath;
      std::stack<char*> tmpStackCurrPath;
      std::stack<char*> tmpStackOutput;
      std::queue<char*> tmpMatchQueue;

      const char* pathSep = "/";

      char* sTmp = strtok(acTmpAbsPath, pathSep);
      while (sTmp)
      {
         tmpStackAbsPath.push(sTmp);
         sTmp = strtok(0, pathSep);
      }

      sTmp = strtok(acTmpCurrDir, pathSep);
      while (sTmp)
      {
         tmpStackCurrPath.push(sTmp);
         sTmp = strtok(0, pathSep);
      }

      sTmp = pcRelPath;
      while (tmpStackCurrPath.size() > tmpStackAbsPath.size())
      {
         *sTmp++ = '.';
         *sTmp++ = '.';
         *sTmp++ = '/';
         tmpStackCurrPath.pop();
      }

      while (tmpStackAbsPath.size() > tmpStackCurrPath.size())
      {
         char* pcTmp = tmpStackAbsPath.top();
         tmpStackOutput.push(pcTmp);
         tmpStackAbsPath.pop();
      }

      while (!tmpStackAbsPath.empty())
      {
         if (strcmp(tmpStackAbsPath.top(),tmpStackCurrPath.top()) == 0)
         {
            tmpMatchQueue.push(tmpStackAbsPath.top());
         }
         else
         {
            while (!tmpMatchQueue.empty())
            {
               tmpStackOutput.push(tmpMatchQueue.front());
               tmpMatchQueue.pop();
            }
            tmpStackOutput.push(tmpStackAbsPath.top());
            sTmp[count++] = '.';
            sTmp[count++] = '.';
            sTmp[count++] = '/';
         }
         tmpStackAbsPath.pop();
         tmpStackCurrPath.pop();
      }
      while (!tmpStackOutput.empty())
      {
         char* pcTmp= tmpStackOutput.top();
         while (*pcTmp != '\0')
         {
            sTmp[count++] = *pcTmp++;
         }
         tmpStackOutput.pop();
         sTmp[count++] = '/';
      }
      sTmp[--count] = '\0';

      relPath = sTmp;
   }*/

   ////////////////////////////////////////////////////////////////////////////////
   FileUtilIOException::FileUtilIOException(const std::string& message, const std::string& filename, unsigned int linenum)
      : dtUtil::Exception(message, filename, linenum)
   {
      mType = &dtUtil::FileExceptionEnum::IOException;
   }

   ////////////////////////////////////////////////////////////////////////////////
   FileNotFoundException::FileNotFoundException(const std::string& message, const std::string& filename, unsigned int linenum)
      : dtUtil::Exception(message, filename, linenum)
   {
      mType = &dtUtil::FileExceptionEnum::FileNotFound;
   }

} // namespace dtUtil
