/* -*-c++-*-
 * allTests - This source file (.h & .cpp) - Using 'The MIT License'
 * Copyright (C) 2005-2008, Alion Science and Technology Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * This software was developed by Alion Science and Technology Corporation under
 * circumstances in which the U. S. Government may have rights in the software.
 *
 * @author William E. Johnson II
 */

#include <prefix/unittestprefix.h>

#include <dtDAL/datatype.h>
#include <dtDAL/exceptionenum.h>

#include <dtUtil/datapathutils.h>
#include <dtUtil/exception.h>
#include <dtUtil/fileutils.h>
#include <dtUtil/log.h>

#include <cppunit/extensions/HelperMacros.h>

#include <osg/Version>
#include <osgDB/FileNameUtils>

class FileUtilsTests : public CPPUNIT_NS::TestFixture
{
   CPPUNIT_TEST_SUITE(FileUtilsTests);

      CPPUNIT_TEST(testFileIO1);
      CPPUNIT_TEST(testFileIO2);
      CPPUNIT_TEST(testRelativePath);
      CPPUNIT_TEST(testConcatPaths);
      //CPPUNIT_TEST(testAbsoluteToRelativePath);
      CPPUNIT_TEST(testDirectoryContentsWithOneFilter);
      CPPUNIT_TEST(testDirectoryContentsWithTwoFilters);
      CPPUNIT_TEST(testDirectoryContentsWithDuplicateFilter);

   CPPUNIT_TEST_SUITE_END();

   public:

      void setUp();
      void tearDown();

      void testFileIO1();
      void testFileIO2();
      void testRelativePath();
      void testConcatPaths();
      void testCopyFileOntoItself();
      //void testAbsoluteToRelativePath();
      void testDirectoryContentsWithOneFilter();
      void testDirectoryContentsWithTwoFilters();
      void testDirectoryContentsWithDuplicateFilter();

   private:

      void NormalizeDirectorySlashes(std::string& str)
      {
         for (size_t i = 0; i < str.size(); i++)
         {
            if (str[i] == '\\')
               str[i] = '/';
         }
      }

      dtUtil::Log* logger;
};

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION(FileUtilsTests);

const std::string DATA_DIR = dtUtil::GetDeltaRootPath()+dtUtil::FileUtils::PATH_SEPARATOR+"examples/data";
const std::string TESTS_DIR = dtUtil::GetDeltaRootPath()+dtUtil::FileUtils::PATH_SEPARATOR+"tests";
const std::string MAPPROJECTCONTEXT = TESTS_DIR + dtUtil::FileUtils::PATH_SEPARATOR + "dtDAL" + dtUtil::FileUtils::PATH_SEPARATOR + "WorkingMapProject";
const std::string PROJECTCONTEXT = TESTS_DIR + dtUtil::FileUtils::PATH_SEPARATOR + "dtDAL" + dtUtil::FileUtils::PATH_SEPARATOR + "WorkingProject";


//////////////////////////////////////////////////////////////////////////
std::string getFileExtensionIncludingDot(const std::string& fileName)
{
#if defined(OSG_VERSION_MAJOR) && defined(OSG_VERSION_MINOR) && OSG_VERSION_MAJOR <= 2  && OSG_VERSION_MINOR <= 6
   //copied from osgDB/FileNameUtils.cpp (v2.8.0)
   std::string::size_type dot = fileName.find_last_of('.');
   if (dot==std::string::npos) return std::string("");
   return std::string(fileName.begin()+dot,fileName.end());
#else
   return osgDB::getFileExtensionIncludingDot(fileName);
#endif
}

//////////////////////////////////////////////////////////////////////////
void FileUtilsTests::setUp()
{
   try
   {
      dtUtil::SetDataFilePathList(dtUtil::GetDataFilePathList() + ";" + dtUtil::GetDeltaRootPath()+"/examples/data/;");
      std::string logName("projectTest");

      logger = &dtUtil::Log::GetInstance(logName);

      dtUtil::FileUtils& fileUtils = dtUtil::FileUtils::GetInstance();
      fileUtils.ChangeDirectory(TESTS_DIR);
      fileUtils.PushDirectory("dtDAL");

      fileUtils.PushDirectory("WorkingProject");
      fileUtils.DirDelete(dtDAL::DataType::STATIC_MESH.GetName(), true);
      fileUtils.DirDelete(dtDAL::DataType::TERRAIN.GetName(), true);
      fileUtils.PopDirectory();

      fileUtils.FileDelete("terrain_simple.ive");
      fileUtils.FileDelete("flatdirt.ive");
      fileUtils.DirDelete("Testing", true);
      fileUtils.DirDelete("recursiveDir", true);

      fileUtils.FileCopy(DATA_DIR + "/models/terrain_simple.ive", ".", false);
      fileUtils.FileCopy(DATA_DIR + "/models/flatdirt.ive", ".", false);
   }
   catch (const dtUtil::Exception& ex)
   {
      CPPUNIT_FAIL(ex.ToString());
   }
}


void FileUtilsTests::tearDown()
{
   dtUtil::FileUtils& fileUtils = dtUtil::FileUtils::GetInstance();

   fileUtils.FileDelete("terrain_simple.ive");
   fileUtils.FileDelete("flatdirt.ive");
   fileUtils.DirDelete("Testing", true);
   fileUtils.DirDelete("recursiveDir", true);

   //Delete a couple other projects
   if (fileUtils.DirExists("TestProject"))
   {
      fileUtils.DirDelete("TestProject", true);
   }
   if (fileUtils.DirExists("Test2Project"))
   {
      fileUtils.DirDelete("Test2Project", true);
   }

   std::string currentDir = fileUtils.CurrentDirectory();
   std::string projectDir("dtDAL");
   if (currentDir.substr(currentDir.size() - projectDir.size()) == projectDir)
   {
      fileUtils.PopDirectory();
   }
}

void FileUtilsTests::testFileIO1()
{
   try
   {
      dtUtil::FileUtils& fileUtils = dtUtil::FileUtils::GetInstance();

      const std::string Dir1("Testing");
      const std::string Dir2Name("Testing1");
      const std::string Dir2(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + Dir2Name);

      const std::string file1("terrain_simple.ive");
      const std::string file2("flatdirt.ive");

      //cleanup
      CPPUNIT_ASSERT_EQUAL_MESSAGE("Deleting a non-existent directory returned as a failure.",
                                    true, fileUtils.DirDelete(Dir1, true));
      fileUtils.MakeDirectory(Dir1);
      fileUtils.MakeDirectory(Dir2);


      dtUtil::FileInfo file1Info = fileUtils.GetFileInfo(file1);
      dtUtil::FileInfo file2Info = fileUtils.GetFileInfo(file2);

      try
      {
         //normally doesn't throw an exception
         fileUtils.GetFileInfo(file2 + "euaoeuaiao.ao.u");
      }
      catch (const dtUtil::Exception& ex)
      {         
         CPPUNIT_FAIL(ex.ToString());
      }

      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should exist.", fileUtils.FileExists(file1));

      fileUtils.FileCopy(file1, Dir1, false);

      CPPUNIT_ASSERT_MESSAGE("The original terrain_simple.ive should exist.", fileUtils.FileExists(file1));

      CPPUNIT_ASSERT_MESSAGE("The new terrain_simple.ive should exist.", fileUtils.FileExists(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      fileUtils.FileCopy(file1, Dir1, true);

      CPPUNIT_ASSERT_MESSAGE("The original terrain_simple.ive should exist.", fileUtils.FileExists(file1));

      CPPUNIT_ASSERT_MESSAGE("The new terrain_simple.ive should exist.", fileUtils.FileExists(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      try
      {
         fileUtils.FileCopy(file2, Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1, false);
         CPPUNIT_FAIL("The file copy should have failed since it was attempting to overwrite the file and overwriting was disabled.");
      }
      catch (const dtUtil::Exception&)
      {
         //correct
      }

      CPPUNIT_ASSERT_MESSAGE("The original flatdirt.ive should exist.", fileUtils.FileExists(file2));

      CPPUNIT_ASSERT_MESSAGE("The new terrain_simple.ive should still exist.", fileUtils.FileExists(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      struct dtUtil::FileInfo fi = fileUtils.GetFileInfo(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1);
      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should be the same size as the original", fi.size == file1Info.size);

      try
      {
         fileUtils.FileCopy(file2, Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1, true);
      }
      catch (const dtUtil::Exception& ex)
      {
         CPPUNIT_FAIL(ex.ToString().c_str());
      }

      CPPUNIT_ASSERT_MESSAGE("The original flatdirt.ive should exist.", fileUtils.FileExists(file2));

      CPPUNIT_ASSERT_MESSAGE("The new terrain_simple.ive, copied from flatdirt.ive, should exist.",
         fileUtils.FileExists(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      fi = fileUtils.GetFileInfo(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1);

      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should be the same size as flatdirt.ive", fi.size == file2Info.size);

   } catch (const dtUtil::Exception& ex)
   {
      CPPUNIT_FAIL(ex.ToString());
   }
}

void FileUtilsTests::testFileIO2()
{
   try
   {
      dtUtil::FileUtils& fileUtils = dtUtil::FileUtils::GetInstance();

      const std::string Dir1("Testing");
      const std::string Dir2Name("Testing1");
      const std::string Dir2(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + Dir2Name);

      const std::string file1("terrain_simple.ive");
      const std::string file2("flatdirt.ive");

      dtUtil::FileInfo file2Info = fileUtils.GetFileInfo(file2);

      //cleanup
      try
      {
         //Shouldn't throw, even if the directory doesn't exist
         fileUtils.DirDelete(Dir1, true);
      }
      catch (const dtUtil::Exception& ex)
      {
         CPPUNIT_FAIL((ex.ToString() + ": Error deleting Directory, but file exists.").c_str());
      }

      fileUtils.MakeDirectory(Dir1);
      fileUtils.MakeDirectory(Dir2);

      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should exist.", fileUtils.FileExists(file1));
      CPPUNIT_ASSERT_MESSAGE("flatdirt.ive should exist.", fileUtils.FileExists(file2));

      try
      {
         fileUtils.FileCopy(file2, Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1, true);
      }
      catch (const dtUtil::Exception& ex)
      {
         CPPUNIT_FAIL(ex.ToString().c_str());
      }

      CPPUNIT_ASSERT_MESSAGE("The original flatdirt.ive should exist.", fileUtils.FileExists(file2));

      CPPUNIT_ASSERT_MESSAGE("The new terrain_simple.ive, copied from flatdirt.ive, should exist.",
         fileUtils.FileExists(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      dtUtil::FileInfo fi = fileUtils.GetFileInfo(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1);

      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should be the same size as flatdirt.ive", fi.size == file2Info.size);

      try
      {
         fileUtils.FileMove(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1, Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1, false);
      }
      catch (const dtUtil::Exception& ex)
      {
         CPPUNIT_FAIL(ex.ToString().c_str());
      }

      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should not exist.",!fileUtils.FileExists(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      CPPUNIT_ASSERT_MESSAGE("The new terrain_simple.ive, copied from terrain_simple.ive, should exist.",
         fileUtils.FileExists(Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      fi = fileUtils.GetFileInfo(Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1);
      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should be the same size as flatdirt.ive", fi.size == file2Info.size);

      //copy the file back so we can try to move it again with overwriting.
      try
      {
         fileUtils.FileCopy(Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1, Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1, false);
      }
      catch (const dtUtil::Exception& ex)
      {
         CPPUNIT_FAIL(ex.ToString().c_str());
      }

      try
      {
         fileUtils.FileMove(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1, Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1, false);
         CPPUNIT_FAIL("Moving the file should have failed since overwriting was turned off.");
      }
      catch (const dtUtil::Exception&)
      {
         //correct
      }

      try
      {
         fileUtils.FileMove(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1, Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1, true);
      }
      catch (const dtUtil::Exception& ex)
      {
         CPPUNIT_FAIL(ex.ToString().c_str());
      }

      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should not exist.",
         !fileUtils.FileExists(Dir1 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      CPPUNIT_ASSERT_MESSAGE("The new terrain_simple.ive, copied from terrain_simple.ive, should exist.",
         fileUtils.FileExists(Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      try
      {
         //copy a directory into itself with "copy contents only"
         fileUtils.DirCopy(Dir1, Dir1+ dtUtil::FileUtils::PATH_SEPARATOR + ".."+ dtUtil::FileUtils::PATH_SEPARATOR + Dir1, true, true);
         CPPUNIT_FAIL("DirCopy should not be able to copy a directory onto itself.");
      }
      catch (const dtUtil::Exception&)
      {
         //correct
      }

      try
      {
         //copy a directory into the parent without contents, so that it would try to recreate the same directory.
         fileUtils.DirCopy(Dir1, ".", true, false);
         CPPUNIT_FAIL("DirCopy should not be able to copy a directory onto itself.");
      }
      catch (const dtUtil::Exception&)
      {
         //correct
      }

      try
      {
         //copy a directory into the parent without contents, so that it would try to recreate the same directory.
         fileUtils.DirCopy(Dir1, Dir1, true, false);
         //doing it again should do nothing.
         fileUtils.DirCopy(Dir1, Dir1, true, false);
      }
      catch (const dtUtil::Exception&)
      {
         CPPUNIT_FAIL("DirCopy should be able to copy a directory into itself as a subdirectory.");
      }

      try
      {
         //Copy a directory into the parent without contents, so that it would try to recreate the same directory.
         //Note that the directory has already been created previously in this same unit test, so this should not
         //be allowed because we'd be overwriting a pre-existing directory.
         fileUtils.DirCopy(Dir1, Dir1, false, false);
         CPPUNIT_FAIL("DirCopy should not be able to overwrite files if overwriting is set to false.");
      }
      catch (const dtUtil::Exception&)
      {
      }

      std::string Dir3("recursiveDir");

      fileUtils.MakeDirectory(Dir3);
      //copy with the directory existing.
      fileUtils.DirCopy(Dir1, Dir3, false);

      CPPUNIT_ASSERT_MESSAGE("The recursive file should exist.",
         fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      CPPUNIT_ASSERT_MESSAGE("The recursive file should exist.",
         fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir1
         + dtUtil::FileUtils::PATH_SEPARATOR + Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));


      fileUtils.DirDelete(Dir3, true);

      CPPUNIT_ASSERT_MESSAGE("The recursive file not should exist.",
         !fileUtils.DirExists(Dir3));

      //copy with the directory NOT existing.
      fileUtils.DirCopy(Dir1, Dir3, false);

      CPPUNIT_ASSERT_MESSAGE("The recursive file should NOT exist.",
         !fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir1
         + dtUtil::FileUtils::PATH_SEPARATOR + Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      CPPUNIT_ASSERT_MESSAGE("The recursive file should exist.",
         fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir2Name + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      CPPUNIT_ASSERT_MESSAGE("The recursive file should exist.",
         fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));


      fileUtils.DirDelete(Dir3, true);

      CPPUNIT_ASSERT_MESSAGE("The recursive file not should exist.",
         !fileUtils.DirExists(Dir3));

      fileUtils.MakeDirectory(Dir3);
      //copy with the directory existing, but with copy contents only. This should be the same as the directory NOT existing before the call.
      fileUtils.DirCopy(Dir1, Dir3, false, true);
      CPPUNIT_ASSERT_MESSAGE("The recursive file should NOT exist.",
         !fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir1
         + dtUtil::FileUtils::PATH_SEPARATOR + Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      CPPUNIT_ASSERT_MESSAGE("The recursive file should exist.",
         fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir2Name + dtUtil::FileUtils::PATH_SEPARATOR + file1));

      CPPUNIT_ASSERT_MESSAGE("The recursive file should exist.",
         fileUtils.FileExists(Dir3 + dtUtil::FileUtils::PATH_SEPARATOR + Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1));


      fi = fileUtils.GetFileInfo(Dir2 + dtUtil::FileUtils::PATH_SEPARATOR + file1);
      CPPUNIT_ASSERT_MESSAGE("terrain_simple.ive should be the same size as flatdirt.ive", fi.size == file2Info.size);

      dtUtil::DirectoryContents dc = fileUtils.DirGetSubs(Dir1);

      CPPUNIT_ASSERT_MESSAGE((Dir1 + " Should only contain 2 entries.").c_str(), dc.size() == 2);

      if (!dc.empty())
      {
         for (dtUtil::DirectoryContents::const_iterator i = dc.begin(); i != dc.end(); ++i)
         {
            const std::string& s = *i;
            CPPUNIT_ASSERT_MESSAGE((Dir1 + " Should only contain 2 entries and they should be \"Testing1\" and \"Testing\".").c_str(),
               s == "Testing1" || s == "Testing");
         }
      }

      //Testing the delete functionality tests DirGetFiles
      try
      {
         CPPUNIT_ASSERT_MESSAGE("Deleting an nonexisten Directory should be ok.", fileUtils.DirDelete("gobbletygook", false) == true);
      }
      catch (const dtUtil::Exception&)
      {
         CPPUNIT_FAIL("Deleting an nonexisten Directory should be ok.");
      }

      //Testing the delete functionality tests DirGetFiles
      try
      {
         fileUtils.DirDelete(Dir1, false);
      }
      catch (const dtUtil::Exception&)
      {
         CPPUNIT_FAIL("Deleting non-empty Directory with a non-recursive call should have returned false.");
      }
      CPPUNIT_ASSERT_MESSAGE(Dir1 + " should still exist.", fileUtils.DirExists(Dir1));
      try
      {
         fileUtils.DirDelete(Dir1, true);
      }
      catch (const dtUtil::Exception& ex)
      {
         CPPUNIT_FAIL((ex.ToString() + ": Deleting non-empty Directory with a non-recursive call should not have generated an Exception.").c_str());
      }
      CPPUNIT_ASSERT_MESSAGE(Dir1 + " should not still exist.", !fileUtils.DirExists(Dir1));
   }
   catch (const dtUtil::Exception& ex)
   {
      CPPUNIT_FAIL(ex.ToString());
   }
}

void FileUtilsTests::testRelativePath()
{
   std::string file = dtUtil::FindFileInPathList("map.xsd");
   CPPUNIT_ASSERT(!file.empty());

   std::string deltaRoot = dtUtil::GetDeltaRootPath();
   CPPUNIT_ASSERT(!deltaRoot.empty());

   deltaRoot = dtUtil::FileUtils::GetInstance().GetAbsolutePath(deltaRoot);

   // Normalize directory separators
   NormalizeDirectorySlashes(file);

   NormalizeDirectorySlashes(deltaRoot);

   std::string relativePath = dtUtil::FileUtils::GetInstance().RelativePath(deltaRoot, file);
   CPPUNIT_ASSERT(!relativePath.empty());

   CPPUNIT_ASSERT_EQUAL_MESSAGE("The relative path should be: data/map.xsd",
         std::string("data/map.xsd"), relativePath);
}

void FileUtilsTests::testConcatPaths()
{
   std::string sepStr(1U, dtUtil::FileUtils::PATH_SEPARATOR);
   CPPUNIT_ASSERT_EQUAL(std::string("test1" + sepStr + "test2.ooo"), dtUtil::FileUtils::ConcatPaths("test1", "test2.ooo"));
   CPPUNIT_ASSERT_EQUAL(std::string("test1" + sepStr + "test2.ooo"), dtUtil::FileUtils::ConcatPaths("test1/", "test2.ooo"));
   //CPPUNIT_ASSERT_EQUAL(std::string("test1" + sepStr + "test2.ooo"), dtUtil::FileUtils::ConcatPaths("test1/", "/test2.ooo"));
   //CPPUNIT_ASSERT_EQUAL(std::string("test1" + sepStr + "test2.ooo"), dtUtil::FileUtils::ConcatPaths("test1\\", "\\test2.ooo"));
   CPPUNIT_ASSERT_EQUAL(std::string("test1" + sepStr + "test2.ooo"), dtUtil::FileUtils::ConcatPaths("test1\\", "test2.ooo"));
   //CPPUNIT_ASSERT_EQUAL(std::string("test1" + sepStr + "test2.ooo"), dtUtil::FileUtils::ConcatPaths("test1", "\\test2.ooo"));
}

void FileUtilsTests::testCopyFileOntoItself()
{
   const std::string testPath = DATA_DIR + "/aNewFilePath";

   const std::string path1 = testPath + "/aFile.txt";
   const std::string path2 = testPath + "/../aNewFilePath/aFile.txt";

   dtUtil::FileUtils::GetInstance().MakeDirectory(testPath);

   FILE* fp = fopen(path1.c_str(), "w");

   CPPUNIT_ASSERT(fp != NULL);

   fputs("Text for my test file", fp);

   fclose(fp);

   dtUtil::FileUtils::GetInstance().FileCopy(path1, path2, true);

   //make sure we didn't blow the file away
   dtUtil::FileInfo fi = dtUtil::FileUtils::GetInstance().GetFileInfo(path1);

   CPPUNIT_ASSERT(fi.size > 0);

   //get rid of test file
   dtUtil::FileUtils::GetInstance().FileDelete(path1);
   //get rid of test directory
   dtUtil::FileUtils::GetInstance().DirDelete(testPath, false);
}

/*void FileUtilsTests::testAbsoluteToRelativePath()
{
   try
   {
      dtUtil::FileUtils &instance = dtUtil::FileUtils::GetInstance();

      std::string path;
      std::string mapXSDPath = dtUtil::FindFileInPathList("map.xsd");

      NormalizeDirectorySlashes(mapXSDPath);

      instance.AbsoluteToRelative(mapXSDPath, path);
      CPPUNIT_ASSERT_EQUAL(std::string("../../data/map.xsd"), path);
   }
   catch(const dtUtil::Exception &e)
   {
      CPPUNIT_FAIL(e.What());
   }
}*/

//////////////////////////////////////////////////////////////////////////
void FileUtilsTests::testDirectoryContentsWithOneFilter()
{
   const dtUtil::DirectoryContents allContents = dtUtil::FileUtils::GetInstance().DirGetFiles(TESTS_DIR);
   const size_t numAllFiles = allContents.size();

   dtUtil::FileExtensionList extensions;
   extensions.push_back(".cpp");

   const dtUtil::DirectoryContents cppContents = dtUtil::FileUtils::GetInstance().DirGetFiles(TESTS_DIR, extensions);
   CPPUNIT_ASSERT_MESSAGE("DirGetFiles() with filtering didn't appear to filter on one extension.",
                          numAllFiles > cppContents.size());

   dtUtil::DirectoryContents::const_iterator itr = cppContents.begin();
   while (itr != cppContents.end())
   {
      CPPUNIT_ASSERT_MESSAGE("DirGetFiles() returned back a file that didn't match the supplied extension",
                             getFileExtensionIncludingDot((*itr)) == ".cpp");
      ++itr;
   }
}

//////////////////////////////////////////////////////////////////////////
void FileUtilsTests::testDirectoryContentsWithTwoFilters()
{
   const dtUtil::DirectoryContents allContents = dtUtil::FileUtils::GetInstance().DirGetFiles(TESTS_DIR);

   dtUtil::FileExtensionList extensions;
   extensions.push_back(".cpp");
   const dtUtil::DirectoryContents cppContents = dtUtil::FileUtils::GetInstance().DirGetFiles(TESTS_DIR, extensions);

   extensions.push_back(".h");
   const dtUtil::DirectoryContents cppHContents = dtUtil::FileUtils::GetInstance().DirGetFiles(TESTS_DIR, extensions);

   //should be more cpp & h files than just cpp files
   CPPUNIT_ASSERT_MESSAGE("DirGetFiles() with filtering didn't appear to filter with two extensions.",
                           cppHContents.size() > cppContents.size());

   dtUtil::DirectoryContents::const_iterator itr = cppHContents.begin();
   while (itr != cppHContents.end())
   {
      CPPUNIT_ASSERT_MESSAGE("DirGetFiles() returned back a file that didn't match the supplied extensions",
         getFileExtensionIncludingDot((*itr)) == ".cpp" ||
         getFileExtensionIncludingDot((*itr)) == ".h");
      ++itr;
   }

}

//////////////////////////////////////////////////////////////////////////
void FileUtilsTests::testDirectoryContentsWithDuplicateFilter()
{
   dtUtil::FileExtensionList singleExtension;
   singleExtension.push_back(".cpp");
   const dtUtil::DirectoryContents singleFilterList = dtUtil::FileUtils::GetInstance().DirGetFiles(TESTS_DIR, singleExtension);

   dtUtil::FileExtensionList duplicateExtension;
   duplicateExtension.push_back(".cpp");
   duplicateExtension.push_back(".cpp"); //add another of the same

   const dtUtil::DirectoryContents duplicateFilter = dtUtil::FileUtils::GetInstance().DirGetFiles(TESTS_DIR, duplicateExtension);

   CPPUNIT_ASSERT_EQUAL_MESSAGE("DirGetFiles() should not count duplicate extensions",
                                 singleFilterList.size(), duplicateFilter.size());
}

